from typing import Any, Iterator

from bigc.api_client import BigCommerceV3APIClient


class BigCommercePricingV3API:
    def __init__(self, api: BigCommerceV3APIClient):
        self._api = api

    def get_pricing(
            self,
            data: dict[str, Any],
            *,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> Iterator[dict[str, Any]]:
        """Return an iterator for batch product pricing"""
        return self._api.post('/pricing/products', data=data, timeout=timeout, retries=retries)



from typing import Any, Iterator

from bigc.api_client import BigCommerceV2APIClient


class BigCommerceCustomerGroupsV2API:
    def __init__(self, api: BigCommerceV2APIClient):
        self._api = api

    def all(
            self,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> Iterator[dict[str, Any]]:
        """Return an iterator for all customer groups"""
        return self._api.get_many('/customer_groups', params=params, timeout=timeout, retries=retries)

    def get(
            self,
            customer_group_id: int,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Get a specific customer group by its ID"""
        return self._api.get(f'/customer_groups/{customer_group_id}', params=params, timeout=timeout, retries=retries)

    def create(self, data: dict[str, Any], *, timeout: float | None = None) -> dict[str, Any]:
        """Create a customer group"""
        return self._api.post('/customer_groups', data=data, timeout=timeout)

    def update(
            self,
            customer_group_id: int,
            data: dict[str, Any],
            *,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Update a specific customer group by its ID"""
        return self._api.put(f'/customer_groups/{customer_group_id}', data=data, timeout=timeout, retries=retries)

    def delete(
            self,
            customer_group_id: int,
            *,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Delete a specific customer group by its ID"""
        return self._api.delete(f'/customer_groups/{customer_group_id}', timeout=timeout, retries=retries)



from typing import Any, Iterator

from bigc.api_client import BigCommerceV3APIClient


class BigCommerceWebhooksV3API:
    def __init__(self, api: BigCommerceV3APIClient):
        self._api = api

    def all(
            self,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> Iterator[dict[str, Any]]:
        """Return an iterator for all webhooks"""
        return self._api.get_many('/hooks', params=params, timeout=timeout, retries=retries)

    def get(self, webhook_id: int, *, timeout: float | None = None, retries: int | None = None) -> dict[str, Any]:
        """Get a specific webhook by its ID"""
        return self._api.get(f'/hooks/{webhook_id}', timeout=timeout, retries=retries)

    def create(
            self,
            data: dict[str, Any],
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
    ) -> dict[str, Any]:
        """Create a webhook under a specific scope"""
        return self._api.post('/hooks', data=data, params=params, timeout=timeout)

    def update(
            self,
            webhook_id: int,
            data: dict[str, Any],
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Update a specific webhook by its ID"""
        return self._api.put(f'/hooks/{webhook_id}', data=data, params=params, timeout=timeout, retries=retries)

    def delete(self, webhook_id: int, *, timeout: float | None = None, retries: int | None = None) -> dict[str, Any]:
        """Delete a specific webhook by its ID"""
        return self._api.delete(f'/hooks/{webhook_id}', timeout=timeout, retries=retries)



from typing import Any, Iterator

from bigc.api_client import BigCommerceV2APIClient


class BigCommerceOrdersV2API:
    def __init__(self, api: BigCommerceV2APIClient):
        self._api = api

    def all(
            self,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> Iterator[dict[str, Any]]:
        """Return an iterator for all orders"""
        return self._api.get_many('/orders', params=params, timeout=timeout, retries=retries)

    def get(
            self,
            order_id: int,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Get an order by its ID"""
        return self._api.get(f'/orders/{order_id}', params=params, timeout=timeout, retries=retries)

    def create(
            self,
            data: dict[str, Any],
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
    ) -> dict[str, Any]:
        """Create an order"""
        return self._api.post('/orders', data=data, params=params, timeout=timeout)

    def update(
            self,
            order_id: int,
            data: dict[str, Any],
            *,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Update a specific order by its ID"""
        return self._api.put(f'/orders/{order_id}', data=data, timeout=timeout, retries=retries)

    def archive(
            self,
            order_id: int,
            *,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> None:
        """Archive a specific order by its ID"""
        self._api.delete(f'/orders/{order_id}', timeout=timeout, retries=retries)

    def all_products(
            self,
            order_id: int,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> Iterator[dict[str, Any]]:
        """Return an iterator for all order products in an order"""
        return self._api.get_many(f'/orders/{order_id}/products', params=params, timeout=timeout, retries=retries)

    def get_product(
            self,
            order_id: int,
            product_id: int,
            *,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Get a specific order product in an order by ID"""
        return self._api.get(f'/orders/{order_id}/products/{product_id}', timeout=timeout, retries=retries)

    def all_shipping_addresses(
            self,
            order_id: int,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> Iterator[dict[str, Any]]:
        """Return an iterator for all order shipping addresses in an order"""
        return self._api.get_many(f'/orders/{order_id}/shipping_addresses', params=params, timeout=timeout, retries=retries)

    def get_shipping_address(
            self,
            order_id: int,
            address_id: int,
            *,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Get a specific shipping address in an order by ID"""
        return self._api.get(f'/orders/{order_id}/shipping_addresses/{address_id}', timeout=timeout, retries=retries)

    def update_shipping_address(
            self,
            order_id: int,
            address_id: int,
            data: dict[str, Any],
            *,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Update a specific shipping address in an order by ID"""
        return self._api.put(f'/orders/{order_id}/shipping_addresses/{address_id}', data=data, timeout=timeout, retries=retries)

    def all_shipments(
            self,
            order_id: int,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> Iterator[dict[str, Any]]:
        """Returns all shipments for a specified order"""
        return self._api.get_many(f'/orders/{order_id}/shipments', params=params, timeout=timeout, retries=retries)

    def get_shipment(
            self,
            order_id: int,
            shipment_id: int,
            *,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Get a shipment by its ID"""
        return self._api.get(f'/orders/{order_id}/shipments/{shipment_id}', timeout=timeout, retries=retries)

    def create_shipment(self, order_id: int, data: dict[str, Any], *, timeout: float | None = None) -> dict[str, Any]:
        """Creates an order shipment for the specified order"""
        return self._api.post(f'/orders/{order_id}/shipments', data=data, timeout=timeout)

    def update_shipment(
            self,
            order_id: int,
            shipment_id: int,
            data: dict[str, Any],
            *,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Updates an order shipment for the specified order"""
        return self._api.put(f'/orders/{order_id}/shipments/{shipment_id}', data=data, timeout=timeout, retries=retries)

    def delete_shipment(
            self,
            order_id: int,
            shipment_id: int,
            *,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> None:
        """Deletes specific shipment by its ID"""
        return self._api.delete(f'/orders/{order_id}/shipments/{shipment_id}', timeout=timeout, retries=retries)

    def all_coupons(
            self,
            order_id: int,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> Iterator[dict[str, Any]]:
        """Return an iterator for all coupons in an order"""
        return self._api.get_many(f'/orders/{order_id}/coupons', params=params, timeout=timeout, retries=retries)



from enum import IntEnum, unique


@unique
class BigCommerceOrderStatus(IntEnum):
    """Enum mapping BigCommerce order statuses to their IDs"""
    INCOMPLETE = 0
    PENDING = 1
    SHIPPED = 2
    PARTIALLY_SHIPPED = 3
    REFUNDED = 4
    CANCELLED = 5
    DECLINED = 6
    AWAITING_PAYMENT = 7
    AWAITING_PICKUP = 8
    AWAITING_SHIPMENT = 9
    COMPLETED = 10
    AWAITING_FULFILLMENT = 11
    MANUAL_VERIFICATION_REQUIRED = 12
    DISPUTED = 13
    PARTIALLY_REFUNDED = 14



from typing import Any, Iterator

from bigc.api_client import BigCommerceV3APIClient
from bigc.exceptions import DoesNotExistError, InvalidDataError


class BigCommerceCustomersV3API:
    def __init__(self, api: BigCommerceV3APIClient):
        self._api = api

    def all(
            self,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> Iterator[dict[str, Any]]:
        """Return an iterator for all customers"""
        return self._api.get_many('/customers', params=params, timeout=timeout, retries=retries, cursor=True)

    def get(
            self,
            customer_id: int,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Get a specific customer by its ID"""
        params = {
            **(params or {}),
            'id:in': customer_id,
        }

        try:
            return self._api.get('/customers', params=params, timeout=timeout, retries=retries)[0]
        except IndexError:
            raise DoesNotExistError() from None

    def create_many(self, data: list[dict[str, Any]], *, timeout: float | None = None) -> list[dict[str, Any]]:
        """Create many customers"""
        return self._api.post('/customers', data=data, timeout=timeout)

    def create(self, data: dict[str, Any], *, timeout: float | None = None) -> dict[str, Any]:
        """Create a single customer"""
        return self.create_many([data], timeout=timeout)[0]

    def update_many(
            self,
            data: list[dict[str, Any]],
            *,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> list[dict[str, Any]]:
        """Update many customers"""
        return self._api.put('/customers', data=data, timeout=timeout, retries=retries)

    def update(
            self,
            customer_id: int,
            data: dict[str, Any],
            *,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Update a single customer"""
        data['id'] = customer_id

        return self.update_many([data], timeout=timeout, retries=retries)[0]

    def delete_many(
            self,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> None:
        """Delete many customers"""
        self._api.delete(f'/customers', params=params, timeout=timeout, retries=retries)

    def delete(
            self,
            customer_id: int,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> None:
        """Delete a single customer"""
        params = {
            **(params or {}),
            'id:in': customer_id,
        }

        self.delete_many(params=params, timeout=timeout, retries=retries)

    def update_form_fields(
            self,
            data: list[dict[str, Any]],
            *,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> list[dict[str, Any]]:
        """Update form-field values"""
        return self._api.put('/customers/form-field-values', data=data, timeout=timeout, retries=retries)

    def update_form_field(
            self,
            customer_id: int,
            data: dict[str, Any],
            *,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Update a single form-field value"""
        return self.update_form_fields([{'customer_id': customer_id, **data}], timeout=timeout, retries=retries)[0]

    def all_addresses(
            self,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> Iterator[dict[str, Any]]:
        """Get all addresses, optionally filtered by a customer's address book"""
        return self._api.get_many(f'/customers/addresses', params=params, timeout=timeout, retries=retries)

    def get_address(
            self,
            address_id: int,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Get one address by its ID, from a customer's address book"""
        params = {
            **(params or {}),
            'id:in': address_id,
        }

        try:
            return self._api.get(f'/customers/addresses', params=params, timeout=timeout, retries=retries)[0]
        except IndexError:
            raise DoesNotExistError() from None

    def create_addresses(self, data: list[dict[str, Any]], *, timeout: float | None = None) -> list[dict[str, Any]]:
        """Create many addresses"""
        return self._api.post('/customers/addresses', data=data, timeout=timeout)

    def create_address(self, data: dict[str, Any], *, timeout: float | None = None) -> dict[str, Any]:
        """Add an address to the customer's address book"""
        try:
            return self.create_addresses([data], timeout=timeout)[0]
        except IndexError:
            raise InvalidDataError('This address already exists.') from None

    def update_addresses(
            self,
            data: list[dict[str, Any]],
            *,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> list[dict[str, Any]]:
        """Update many addresses"""
        return self._api.put('/customers/addresses', data=data, timeout=timeout, retries=retries)

    def update_address(
            self,
            address_id: int,
            data: dict[str, Any],
            *,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Update an address by its ID"""
        try:
            return self.update_addresses([{'id': address_id, **data}], timeout=timeout, retries=retries)[0]
        except IndexError:
            raise InvalidDataError('This address already exists.') from None

    def delete_addresses(
            self,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> None:
        """Delete many addresses"""
        self._api.delete(f"/customers/addresses", params=params, timeout=timeout, retries=retries)

    def delete_address(
            self,
            address_id: int,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> None:
        """Delete an address by its ID"""
        params = {
            **(params or {}),
            'id:in': address_id,
        }

        self.delete_addresses(params=params, timeout=timeout, retries=retries)



from __future__ import annotations

import inspect
from typing import Any, ClassVar

_status_code_exc_class_map: dict[int, type[BigCommerceException]] = {}


class BigCommerceException(Exception):
    """Base exception class for BigCommerce API errors."""

    DEFAULT_MESSAGE: ClassVar[str] = 'Request to BigCommerce failed.'
    STATUS_CODE: ClassVar[int | None] = None

    def __init__(self,
                 message: str | None = None,
                 status_code: int | None = None,
                 response: Any = None,
                 errors: dict[str, str] | None = None):
        """
        :param message: A short description of the error's cause.
        :param status_code: The HTTP status code of the API response.
        :param response: The underlying API response, if any, that caused the
            error. Currently, this is a ``requests.Response``, but this is an
            implementation detail that's subject to change.
        :param errors: A dictionary that may contain additional error details.
        """
        super().__init__(message, status_code, response, errors)

        self.message = message
        self.status_code = status_code
        self.response = response
        self.errors = errors

    def __str__(self):
        return self.message or self.DEFAULT_MESSAGE

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        # Only include subclasses defined in this file
        if inspect.getmodule(cls).__name__ == __name__:
            if cls.STATUS_CODE is not None:
                _status_code_exc_class_map[cls.STATUS_CODE] = cls

    @staticmethod
    def get_exc_class_for_status_code(status_code: int) -> type[BigCommerceException]:
        try:
            return _status_code_exc_class_map[status_code]
        except KeyError:
            if 400 <= status_code < 500:
                return BigCommerceClientError
            if 500 <= status_code < 600:
                return BigCommerceServerError
            return BigCommerceException

    @staticmethod
    def extract_error_message(response_data: object) -> tuple[str | None, dict[str, str] | None]:
        message: str | None = None
        errors: dict[str, str] | None = None

        # Object-style error messages
        if isinstance(response_data, dict):
            message = response_data.get('title')
            errors = response_data.get('errors')
        # Array-style error messages
        elif isinstance(response_data, list) and len(response_data) > 0:
            if isinstance(response_data[0], dict):
                message = response_data[0].get('message')

        return message, errors


class BigCommerceClientError(BigCommerceException):
    """Exception class for 4xx errors from the BigCommerce API."""


class BadRequestError(BigCommerceException):
    DEFAULT_MESSAGE = 'The request was malformed in some way.'
    STATUS_CODE = 400


class UnauthorizedError(BigCommerceClientError):
    DEFAULT_MESSAGE = 'Credentials were not provided or are invalid.'
    STATUS_CODE = 401


class ForbiddenError(BigCommerceClientError):
    DEFAULT_MESSAGE = 'The request had insufficient permissions for this operation.'
    STATUS_CODE = 403


class DoesNotExistError(BigCommerceClientError):
    DEFAULT_MESSAGE = 'The requested entity does not exist.'
    STATUS_CODE = 404


class ConflictError(BigCommerceClientError):
    DEFAULT_MESSAGE = 'The requested change has been rejected by BigCommerce.'
    STATUS_CODE = 409


class EntityTooLargeError(BigCommerceClientError):
    DEFAULT_MESSAGE = 'Too many objects were requested.'
    STATUS_CODE = 413


class InvalidDataError(BigCommerceClientError):
    DEFAULT_MESSAGE = 'The request omitted required fields or contained invalid data.'
    STATUS_CODE = 422


class LockedError(BigCommerceClientError):
    DEFAULT_MESSAGE = 'The requested resource is currently unavailable.'
    STATUS_CODE = 423


class TooManyRequestsError(BigCommerceClientError):
    DEFAULT_MESSAGE = "The store's rate limit has been exceeded."
    STATUS_CODE = 429


class BigCommerceServerError(BigCommerceException):
    """Exception class for 5xx errors from the BigCommerce API."""


class InternalServerError(BigCommerceServerError):
    DEFAULT_MESSAGE = 'An error has occurred within the BigCommerce server.'
    STATUS_CODE = 500


class BadGatewayError(BigCommerceServerError):
    DEFAULT_MESSAGE = 'Bad gateway.'
    STATUS_CODE = 502


class ServiceUnavailableError(BigCommerceServerError):
    DEFAULT_MESSAGE = 'The store is currently offline or suspended.'
    STATUS_CODE = 503


class GatewayTimeoutError(BigCommerceServerError):
    DEFAULT_MESSAGE = 'The request has timed out.'
    STATUS_CODE = 504


class InsufficientStorageError(BigCommerceServerError):
    DEFAULT_MESSAGE = "The store has exceeded its plan's limit for a resource."
    STATUS_CODE = 507


class BigCommerceNetworkError(BigCommerceException):
    """Exception class for network-related errors."""


__all__ = (
    'BigCommerceException',
    'BigCommerceClientError',
    'BadRequestError',
    'UnauthorizedError',
    'ForbiddenError',
    'DoesNotExistError',
    'ConflictError',
    'EntityTooLargeError',
    'InvalidDataError',
    'LockedError',
    'TooManyRequestsError',
    'BigCommerceServerError',
    'InternalServerError',
    'BadGatewayError',
    'ServiceUnavailableError',
    'GatewayTimeoutError',
    'InsufficientStorageError',
    'BigCommerceNetworkError',
)



from typing import Any, Iterator

from bigc.api_client import BigCommerceV2APIClient


class BigCommerceCurrenciesV2API:
    def __init__(self, api: BigCommerceV2APIClient):
        self._api = api

    def all(
            self,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> Iterator[dict[str, Any]]:
        """Return an iterator for all currencies"""
        return self._api.get_many('/currencies', params=params, timeout=timeout, retries=retries)

    def get(
            self,
            currency_id: int,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ):
        """Get a specific currency by its ID"""
        return self._api.get(f'/currencies/{currency_id}', params=params, timeout=timeout, retries=retries)



"""
Resources are organized by their URLs and API version. For example, all v3 endpoints that start
with `/customers` reside within `customers_v3.py`.
"""

from .carts_v3 import BigCommerceCartsV3API
from .categories_v3 import BigCommerceCategoriesV3API
from .checkouts_v3 import BigCommerceCheckoutsV3API
from .currencies_v2 import BigCommerceCurrenciesV2API
from .customer_groups_v2 import BigCommerceCustomerGroupsV2API
from .customers_v3 import BigCommerceCustomersV3API
from .orders_v2 import BigCommerceOrdersV2API
from .orders_v3 import BigCommerceOrdersV3API
from .pricing_v3 import BigCommercePricingV3API
from .product_variants_v3 import BigCommerceProductVariantsV3API
from .products_v3 import BigCommerceProductsV3API
from .webhooks_v3 import BigCommerceWebhooksV3API



import itertools
from abc import ABC, abstractmethod
from typing import Any, Iterator, NoReturn

import requests

from bigc.exceptions import (
    BadGatewayError,
    BigCommerceException,
    BigCommerceNetworkError,
    GatewayTimeoutError,
    InternalServerError,
    ServiceUnavailableError,
)

MAX_V2_PAGE_SIZE = 250
MAX_V3_PAGE_SIZE = 250


class BigCommerceRequestClient(ABC):
    def __init__(
            self,
            store_hash: str,
            access_token: str,
            *,
            timeout: float | None = None,
            get_retries: int | None = None
    ):
        self.store_hash = store_hash
        self.access_token = access_token
        self.timeout = timeout
        self.get_retries = get_retries

    def request(
            self,
            method: str,
            path: str,
            *,
            data: Any = None,
            params: dict[str, Any] | None = None,
            headers: dict[str, str] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> Any:
        method = method.upper()

        if headers is None:
            headers = {}
        if timeout is None:
            timeout = self.timeout
        if retries is None:
            if method == 'GET':
                retries = self.get_retries

            retries = retries or 0

        self._validate_path(path)
        self._validate_retries(method, retries)

        url = self._prepare_url(path)
        params = self._process_params(params)
        headers = self._get_standard_request_headers() | headers

        def perform_request() -> Any:
            try:
                response = requests.request(
                    method,
                    url,
                    json=data,
                    params=params,
                    headers=headers,
                    timeout=timeout,
                )
            except requests.Timeout as exc:
                raise GatewayTimeoutError() from exc
            except requests.RequestException as exc:
                raise BigCommerceNetworkError() from exc

            if response.ok:
                # Return None for empty responses instead of raising
                return response.json() if response.text else None
            else:
                self._handle_error_response(response)

        last_exc: Exception | None = None
        for _ in range(retries + 1):
            try:
                return perform_request()
            except (
                    InternalServerError,
                    BadGatewayError,
                    ServiceUnavailableError,
                    GatewayTimeoutError,
                    BigCommerceNetworkError,
            ) as exc:
                last_exc = exc

        raise last_exc

    def get(self, *args, **kwargs):
        """Alias for ``request('GET', ...)``"""
        return self.request('GET', *args, **kwargs)

    def post(self, *args, **kwargs):
        """Alias for ``request('POST', ...)``"""
        return self.request('POST', *args, **kwargs)

    def put(self, *args, **kwargs):
        """Alias for ``request('PUT', ...)``"""
        return self.request('PUT', *args, **kwargs)

    def delete(self, *args, **kwargs):
        """Alias for ``request('DELETE', ...)``"""
        return self.request('DELETE', *args, **kwargs)

    @abstractmethod
    def get_many(
            self,
            path: str,
            *,
            page_size: int | None = None,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> Iterator[Any]:
        """Make a request to a paginated BigCommerce API endpoint"""
        pass

    @abstractmethod
    def _prepare_url(self, path: str) -> str:
        pass

    @staticmethod
    def _validate_path(path: str) -> None:
        if '?' in path:
            raise ValueError('path should not contain query parameters')

        if '#' in path:
            raise ValueError('path should not contain fragment')

    @staticmethod
    def _validate_retries(method: str, retries: int) -> None:
        if retries < 0:
            raise ValueError('retries must be 0 or greater')

        if method == 'POST' and retries:
            raise ValueError('POST requests cannot be safely retried')

    def _get_standard_request_headers(self) -> dict[str, str]:
        return {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'X-Auth-Token': self.access_token,
        }

    @staticmethod
    def _process_params(params: dict[str, Any] | None) -> dict[str, str] | None:
        if not params:
            return None

        def _process_param(value: Any) -> str:
            if isinstance(value, list | tuple | set):
                return ','.join(map(str, value))

            return str(value)

        return {k: _process_param(v) for k, v in params.items()}

    @staticmethod
    def _handle_error_response(response: requests.Response) -> NoReturn:
        try:
            message, errors = BigCommerceException.extract_error_message(response.json())
        except requests.JSONDecodeError:
            message, errors = None, None

        exc_class = BigCommerceException.get_exc_class_for_status_code(response.status_code)
        raise exc_class(
            message=message,
            status_code=response.status_code,
            response=response,
            errors=errors,
        )


class BigCommerceV2APIClient(BigCommerceRequestClient):
    """A client for directly calling BigCommerce v2 API endpoints"""

    def _prepare_url(self, path: str) -> str:
        return f"https://api.bigcommerce.com/stores/{self.store_hash}/v2/{path.lstrip('/')}"

    def get_many(
            self,
            path: str,
            *,
            page_size: int | None = None,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> Iterator[Any]:
        page_size = MAX_V2_PAGE_SIZE if page_size is None else int(page_size)

        params = {**params} if params else {}

        if params.keys() & {'limit', 'offset'}:
            raise ValueError('params already has pagination values (limit and/or offset)')

        params['limit'] = page_size

        for cur_page in itertools.count(1):
            params['page'] = cur_page

            res_data = super().get(path, params=params, timeout=timeout, retries=retries)

            # The API returns HTTP 204 (empty) past the last page
            if res_data is None:
                return

            if not isinstance(res_data, list):
                raise TypeError(f'expected list, got {type(res_data).__name__}')

            yield from res_data

            # Check if we're on the last page
            if len(res_data) < page_size:
                return


class BigCommerceV3APIClient(BigCommerceRequestClient):
    """A client for directly calling BigCommerce v3 API endpoints"""

    def _prepare_url(self, path: str) -> str:
        return f"https://api.bigcommerce.com/stores/{self.store_hash}/v3/{path.lstrip('/')}"

    def request(self, *args, **kwargs):
        # v3 response bodies are boxed in the 'data' key
        response = super().request(*args, **kwargs)
        return None if response is None else response['data']

    def _get_many_using_limit_offset(
            self,
            path: str,
            *,
            page_size: int,
            params: dict[str, Any],
            timeout: float | None,
            retries: int | None = None,
    ) -> Iterator[Any]:
        if params.keys() & {'limit', 'offset'}:
            raise ValueError('params already has pagination values (limit and/or offset)')

        params['limit'] = page_size

        cur_page = 1
        num_pages = 1  # Will be set to the right value in the loop
        while cur_page <= num_pages:
            params['page'] = cur_page

            res_data = super().request('GET', path, params=params, timeout=timeout, retries=retries)

            cur_page += 1
            num_pages = int(res_data['meta']['pagination']['total_pages'])

            if not isinstance(res_data['data'], list):
                raise TypeError(f"expected list, got {type(res_data['data']).__name__}")

            yield from res_data['data']

    def _get_many_using_cursor(
            self,
            path: str,
            *,
            page_size: int,
            params: dict[str, Any],
            timeout: float | None,
            retries: int | None = None,
    ) -> Iterator[Any]:
        if params.keys() & {'limit', 'before', 'after'}:
            raise ValueError('params already has pagination values (limit, before, and/or after)')

        params['limit'] = page_size

        while True:
            response = super().request('GET', path, params=params, timeout=timeout, retries=retries)

            yield from response['data']

            if not (
                    # end_cursor will still be provided if the next page is empty
                    response['meta']['cursor_pagination']['links'].get('next')
                    and response['data']
            ):
                break

            params['after'] = response['meta']['cursor_pagination']['end_cursor']

    def get_many(
            self,
            path: str,
            *,
            page_size: int | None = None,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
            cursor: bool = False,
    ) -> Iterator[Any]:
        page_size = MAX_V3_PAGE_SIZE if page_size is None else int(page_size)
        params = {**params} if params else {}

        if cursor:
            return self._get_many_using_cursor(
                path,
                page_size=page_size,
                params=params,
                timeout=timeout,
                retries=retries,
            )
        else:
            return self._get_many_using_limit_offset(
                path,
                page_size=page_size,
                params=params,
                timeout=timeout,
                retries=retries,
            )



from typing import Any, Iterator

from bigc.api_client import BigCommerceV3APIClient


class BigCommerceProductsV3API:
    def __init__(self, api: BigCommerceV3APIClient):
        self._api = api

    def all(
            self,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> Iterator[dict[str, Any]]:
        """Return an iterator for all products"""
        return self._api.get_many('/catalog/products', params=params, timeout=timeout, retries=retries)

    def get(
            self,
            product_id: int,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Get a specific product by its ID"""
        return self._api.get(f'/catalog/products/{product_id}', params=params, timeout=timeout, retries=retries)

    def create(
            self,
            data: dict[str, Any],
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
    ) -> dict[str, Any]:
        """Create a product"""
        return self._api.post('/catalog/products', data=data, params=params, timeout=timeout)

    def update(
            self,
            product_id: int,
            data: dict[str, Any],
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Update a specific product by its ID"""
        return self._api.put(f'/catalog/products/{product_id}', data=data, params=params, timeout=timeout, retries=retries)

    def delete(self, product_id: int, *, timeout: float | None = None, retries: int | None = None) -> dict[str, Any]:
        """Delete a specific product by its ID"""
        return self._api.delete(f'/catalog/products/{product_id}', timeout=timeout, retries=retries)



from typing import Any, TypeAlias
from uuid import UUID

from bigc.api_client import BigCommerceV3APIClient

UUIDLike: TypeAlias = UUID | str


class BigCommerceCheckoutsV3API:
    def __init__(self, api: BigCommerceV3APIClient):
        self._api = api

    def get(
            self,
            checkout_id: UUIDLike,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Get a specific checkout by its ID"""
        return self._api.get(f'/checkouts/{checkout_id}', params=params, timeout=timeout, retries=retries)

    def update(
            self,
            checkout_id: UUIDLike,
            data: dict[str, Any],
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Change customer message pertaining to an existing Checkout"""
        return self._api.put(f'/checkouts/{checkout_id}', data=data, params=params, timeout=timeout, retries=retries)

    def add_billing_address(
            self,
            checkout_id: UUIDLike,
            data: dict[str, Any],
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
    ) -> dict[str, Any]:
        """Add a billing address to an existing checkout"""
        return self._api.post(f'/checkouts/{checkout_id}/billing-address', data=data, params=params, timeout=timeout)

    def update_billing_address(
            self,
            checkout_id: UUIDLike,
            address_id: str,
            data: dict[str, Any],
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Update an existing billing address on a checkout"""
        return self._api.put(f'/checkouts/{checkout_id}/billing-address/{address_id}', data=data, params=params, timeout=timeout, retries=retries)

    def add_consignments(
            self,
            checkout_id: UUIDLike,
            data: list[dict[str, Any]],
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
    ) -> dict[str, Any]:
        """Add a new consignment to a checkout"""
        return self._api.post(f'/checkouts/{checkout_id}/consignments', data=data, params=params, timeout=timeout)

    def update_consignment(
            self,
            checkout_id: UUIDLike,
            consignment_id: str,
            data: dict[str, Any],
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Update an existing consignment's selected shipping option"""
        return self._api.put(f'/checkouts/{checkout_id}/consignments/{consignment_id}', data=data, params=params, timeout=timeout, retries=retries)

    def delete_consignment(
            self,
            checkout_id: UUIDLike,
            consignment_id: str,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Remove an existing consignment from a checkout"""
        return self._api.delete(f'/checkouts/{checkout_id}/consignments/{consignment_id}', params=params, timeout=timeout, retries=retries)

    def add_coupon(
            self,
            checkout_id: UUIDLike,
            data: dict[str, Any],
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
    ) -> dict[str, Any]:
        """Add a coupon code to a checkout"""
        return self._api.post(f'/checkouts/{checkout_id}/coupons', data=data, params=params, timeout=timeout)

    def delete_coupon(
            self,
            checkout_id: UUIDLike,
            coupon_code: str,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Delete a coupon code from a checkout"""
        return self._api.delete(f'/checkouts/{checkout_id}/coupons/{coupon_code}', params=params, timeout=timeout, retries=retries)

    def add_discounts(
            self,
            checkout_id: UUIDLike,
            data: dict[str, Any],
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
    ) -> dict[str, Any]:
        """Add discounts to an existing checkout"""
        return self._api.post(f'/checkouts/{checkout_id}/discounts', data=data, params=params, timeout=timeout)

    def create_order(
            self,
            checkout_id: UUIDLike,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
    ) -> dict[str, Any]:
        """Create an order"""
        return self._api.post(f'/checkouts/{checkout_id}/orders', params=params, timeout=timeout)



from typing import Any, TypeAlias
from uuid import UUID

from bigc.api_client import BigCommerceV3APIClient

UUIDLike: TypeAlias = UUID | str


class BigCommerceCartsV3API:
    def __init__(self, api: BigCommerceV3APIClient):
        self._api = api

    def get(
            self,
            cart_id: UUIDLike,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Get a specific cart by its ID"""
        return self._api.get(f'/carts/{cart_id}', params=params, timeout=timeout, retries=retries)

    def create(
            self,
            data: dict[str, Any],
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
    ) -> dict[str, Any]:
        """Create a new cart"""
        return self._api.post('/carts', data=data, params=params, timeout=timeout)

    def update(
            self,
            cart_id: UUIDLike,
            data: dict[str, Any],
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Update a specific cart by its ID"""
        return self._api.put(f'/carts/{cart_id}', data=data, params=params, timeout=timeout, retries=retries)

    def delete(self, cart_id: UUIDLike, *, timeout: float | None = None, retries: int | None = None) -> None:
        """Delete a specific cart by its ID"""
        self._api.delete(f'/carts/{cart_id}', timeout=timeout, retries=retries)

    def add_line_items(
            self,
            cart_id: UUIDLike,
            data: dict[str, Any],
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
    ) -> dict[str, Any]:
        """Add line items to a cart"""
        return self._api.post(f'/carts/{cart_id}/items', data=data, params=params, timeout=timeout)

    def update_line_item(
            self,
            cart_id: UUIDLike,
            item_id: UUIDLike,
            data: dict[str, Any],
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Update a single line item in a cart"""
        return self._api.put(f'/carts/{cart_id}/items/{item_id}', data=data, params=params, timeout=timeout, retries=retries)

    def delete_line_item(
            self,
            cart_id: UUIDLike,
            item_id: UUIDLike,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Remove a line item from a cart"""
        return self._api.delete(f'/carts/{cart_id}/items/{item_id}', params=params, timeout=timeout, retries=retries)

    def create_redirect_url(
            self,
            cart_id: UUIDLike,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
    ) -> dict[str, Any]:
        """Create a redirect URL for a cart"""
        return self._api.post(f'/carts/{cart_id}/redirect_urls', params=params, timeout=timeout)



from bigc.api_client import BigCommerceV2APIClient, BigCommerceV3APIClient
from bigc.resources import *


class BigCommerceAPI:
    def __init__(
            self,
            store_hash: str,
            access_token: str,
            *,
            timeout: float | None = None,
            get_retries: int | None = None
    ):
        api_v2 = BigCommerceV2APIClient(store_hash, access_token, timeout=timeout, get_retries=get_retries)
        api_v3 = BigCommerceV3APIClient(store_hash, access_token, timeout=timeout, get_retries=get_retries)

        self.api_v2 = api_v2
        self.api_v3 = api_v3

        self.carts_v3: BigCommerceCartsV3API = BigCommerceCartsV3API(api_v3)
        self.categories_v3: BigCommerceCategoriesV3API = BigCommerceCategoriesV3API(api_v3)
        self.checkouts_v3: BigCommerceCheckoutsV3API = BigCommerceCheckoutsV3API(api_v3)
        self.currencies_v2: BigCommerceCurrenciesV2API = BigCommerceCurrenciesV2API(api_v2)
        self.customer_groups_v2: BigCommerceCustomerGroupsV2API = BigCommerceCustomerGroupsV2API(api_v2)
        self.customers_v3: BigCommerceCustomersV3API = BigCommerceCustomersV3API(api_v3)
        self.orders_v2: BigCommerceOrdersV2API = BigCommerceOrdersV2API(api_v2)
        self.orders_v3: BigCommerceOrdersV3API = BigCommerceOrdersV3API(api_v3)
        self.pricing_v3: BigCommercePricingV3API = BigCommercePricingV3API(api_v3)
        self.product_variants_v3: BigCommerceProductVariantsV3API = BigCommerceProductVariantsV3API(api_v3)
        self.products_v3: BigCommerceProductsV3API = BigCommerceProductsV3API(api_v3)
        self.webhooks_v3: BigCommerceWebhooksV3API = BigCommerceWebhooksV3API(api_v3)



from datetime import datetime, timezone
from email.utils import mktime_tz, parsedate_tz

__all__ = ('parse_rfc2822_date',)


# The BigCommerce API uses RFC-2822 sometimes, and there isn't a one-liner
# in the standard library to parse it.
def parse_rfc2822_date(date_str: str) -> datetime:
    """Parse an RFC-2822 date using ``email.utils.parsedate_tz`` and return a ``datetime`` instance"""
    parsed_parts: tuple = parsedate_tz(date_str)
    utc_timestamp: int = mktime_tz(parsed_parts)
    return datetime.fromtimestamp(utc_timestamp, tz=timezone.utc)



from typing import Any, Iterator

from bigc.api_client import BigCommerceV3APIClient
from bigc.exceptions import DoesNotExistError


class BigCommerceOrdersV3API:
    def __init__(self, api: BigCommerceV3APIClient):
        self._api = api

    def get_refund_quote(
            self,
            order_id: int,
            data: dict[str, Any],
            *,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Get a refund quote for an order by ID"""
        return self._api.post(f'/orders/{order_id}/payment_actions/refund_quotes', data=data, timeout=timeout, retries=retries)

    def create_refund(
            self,
            order_id: int,
            data: dict[str, Any],
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
    ) -> dict[str, Any]:
        """Create a refund for an order by ID"""
        return self._api.post(f'/orders/{order_id}/payment_actions/refunds', data=data, params=params, timeout=timeout)

    def all_refunds(
            self,
            order_id: int | None = None,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> Iterator[dict[str, Any]]:
        """Return an iterator for all refunds, optionally filtered by order"""
        if order_id:
            endpoint = f'/orders/{order_id}/payment_actions/refunds'
        else:
            endpoint = '/orders/payment_actions/refunds'

        return self._api.get_many(endpoint, params=params, timeout=timeout, retries=retries)

    def get_refund(
            self,
            refund_id: int,
            *,
            params: dict[str, Any] | None = None,
            timeout: float | None = None,
            retries: int | None = None,
    ) -> dict[str, Any]:
        """Get a specific refund by its ID"""
        params = {
            **(params or {}),
            'id:in': refund_id,
        }

        try:
            return self._api.get('/orders/payment_actions/refunds', params=params, timeout=timeout, retries=retries)[0]
        except IndexError:
            raise DoesNotExistError() from None
